<?php

namespace Drupal\uswds_storybook\Element;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Attribute\FormElement;
use Drupal\Core\Render\Element\FormElementBase;

/**
 * Form element to enter approximate time.
 */
#[FormElement('approximate_time')]
class ApproximateTime extends FormElementBase {

  const TIME_FORMAT_REGEXP = '!^([0-2][0-9]):([0-5][05])(:00)?$!';

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = static::class;
    return [
      '#input' => TRUE,
      '#increment' => 30,
      '#min' => '00:00',
      '#max' => '23:59',
      '#process' => [
        [$class, 'processApproximateTime'],
      ],
      '#element_validate' => [
        [$class, 'validateApproximateTime'],
      ],
      '#theme' => 'input__textfield',
      '#theme_wrappers' => [
        'form_element' => [
          '#attributes' => [
            'class' => ['usa-form-group'],
          ],
        ],
      ],
      '#attached' => [
        'library' => [
          'uswds_storybook/time_picker',
        ]
      ],
    ];
  }

  /**
   * Process callback for approximate_time.
   */
  public static function processApproximateTime(array &$element, FormStateInterface $form_state, &$complete_form): array {
    $element['#attributes']['class'][] = 'usa-input';
    $described_by = [$element['#id'] . '-label'];
    if (!empty($element['#description'])) {
      $described_by[] = $element['#id'] . '--description';
    }
    $element['#attributes']['aria-describedby'] = implode(' ', $described_by);
    $element['#prefix'] = '<div class="usa-time-picker">';
    $element['#suffix'] = '</div>';

    self::setTimeAttribute($element, 'increment', 'data-step');
    self::setTimeAttribute($element, 'min', 'data-min-time');
    self::setTimeAttribute($element, 'max', 'data-max-time');

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state): ?string {
    if ($input === FALSE && !empty($element['#default_value'])) {
      return $element['#default_value'];
    }
    if ($input !== FALSE && $input !== NULL) {
      if (strlen($input) == 5) {
        $input .= ':00';
      }
      $pseudo_element = [
        '#value' => $input,
        '#min' => $element['#min'] ?? '00:00',
        '#max' => $element['#max'] ?? '23:59',
        '#preprocessed_min_max' => TRUE,
      ];
      if (self::isTimeValueValid($pseudo_element)) {
        return $input;
      }
    }
    return NULL;
  }

  /**
   * Validate callback for approximate_time.
   */
  public static function validateApproximateTime(&$element, FormStateInterface $form_state, &$complete_form): void {
    if (!self::isTimeValueValid($element)) {
      $form_state->setError($element, t('%name field is not within bounds.', ['%name' => $element['#title']]));
    }
  }

  /**
   * Determines if a time value is properly formed and within bounds.
   *
   * @param string|array $value
   *   Either the value to be checked, or an element wrapping the value.
   *
   * @return bool
   *   TRUE if the value is valid, FALSE otherwise.
   */
  protected static function isTimeValueValid(array $element): bool {
    $value = $element['#value'] ?? '';
    // Objects and arrays are automatically invalid.
    if (!is_scalar($value)) {
      return FALSE;
    }
    // Booleans are automatically invalid.
    if (is_bool($value)) {
      return FALSE;
    }
    // Empty string is always considered valid. If the field is required, a
    // separate validator will catch that.
    if ($value === '') {
      return TRUE;
    }

    // If string is formatted as HH:MM, and hour is between 0 and 23, return
    // TRUE.
    if (preg_match(self::TIME_FORMAT_REGEXP, $value, $matches)) {
      $hour = intval($matches[1]);
      $minute = intval(ltrim($matches[2], '0'));
      $increment = intval($element['#incremenet'] ?? 30);
      if ($hour > 23) {
        // Invalid hour.
        return FALSE;
      }
      // Short-circuit if we've already determined min/max.
      if (isset($element['#preprocessed_min_max'])) {
        $min = $element['#min'];
        $max = $element['#max'];
      }
      else {
        // Nope, gotta do it the long way.
        // Since we'll be doing the same calculation twice, we can create an
        // anonymous function for DRY purposes.
        $get_value = function(array $element, string $key, string $default): string {
          if (isset($element[$key]) && preg_match(self::TIME_FORMAT_REGEXP, $element[$key])) {
            return $element[$key];
          }
          return $default;
        };

        $min = $get_value($element, '#min', '00:00');
        $max = $get_value($element, '#max', '23:59');
      }
      if (
        $value >= $min
        && $value <= $max
        // Ensure that we are on an increment-valid time.
        && (($increment == 60 && $minute == 0) || $minute % $increment === 0)
      ) {
        // We are within min/max values, and therefore valid.
        return TRUE;
      }
    }
    // Nope, failed all tests.
    return FALSE;
  }

  /**
   * Helper function to set a data-attribute on the input element.
   *
   * @param array $element
   *   The element descriptor.
   * @param string $attr_name
   *   The name of the attribute to find.
   * @param string|null $data_name
   *   The name of the attribute to replace it with.
   *
   * @return void
   */
  protected static function setTimeAttribute(array &$element, string $attr_name, ?string $data_name = NULL): void {
    if (isset($element["#$attr_name"])) {
      $attr_val = $element["#$attr_name"];
    }
    elseif (isset($element['#attributes'][$attr_name])) {
      $attr_val = $element['#attributes'][$attr_name];
      unset($element['#attributes'][$attr_name]);
    }
    else {
      return;
    }
    $element['#attributes'][$data_name ?? "data-$attr_name"] = $attr_val;
  }

}
