<?php

namespace Drupal\uswds_storybook\Element;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Attribute\FormElement;
use Drupal\Core\Render\Element\FormElementBase;

#[FormElement('memorable_date')]
class MemorableDate extends FormElementBase {

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = static::class;
    return [
      '#input' => TRUE,
      '#process' => [
        [$class, 'processMemorableDate'],
      ],
      '#element_validate' => [
        [$class, 'validateMemorableDate'],
      ],
      '#theme_wrappers' => ['fieldset'],
      '#attached' => [
        'library' => [
          'uswds_storybook/memorable_date',
        ]
      ],
      '#prefix' => '<div class="usa-memorable-date">',
      '#suffix' => '</div>',
      '#min' => '1900-01-01',
      '#max' => '2038-01-09',
    ];
  }

  /**
   * Process callback for this element.
   */
  public static function processMemorableDate(&$element, FormStateInterface $form_state, &$form): array {
    $default_value = $element['#default_value'];
    $default_values = ['year' => NULL, 'month' => NULL, 'day' => NULL];
    if (preg_match('/^(?P<year>[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})/', $default_value, $matches)) {
      $default_values = array_filter($matches, fn($key) => !is_numeric($key), ARRAY_FILTER_USE_KEY);
      $default_values = array_map(fn($item) => intval($item), $default_values);
    }
    $id = $element['#id'];
    $full_default_value = implode('-', $default_values);
    $parts = NULL;
    if (!empty($element['#min']) && $full_default_value < $element['#min']) {
      $parts = explode('-', $element['#min']);
    }
    elseif (!empty($element['#max']) && $full_default_value > $element['#max']) {
      $parts = explode('-', $element['#max']);
    }
    if (!empty($parts)) {
      $default_values = [
        'year' => intval($parts[0]),
        'month' => intval($parts[1]),
        'day' => intval($parts[2]),
      ];
    }

    $element['month'] = [
      '#type' => 'select',
      '#title' => t('Month'),
      '#options' => [
        "" => t('- Select -'),
        '1' => t('January'),
        '2' => t('February'),
        '3' => t('March'),
        '4' => t('April'),
        '5' => t('May'),
        '6' => t('June'),
        '7' => t('July'),
        '8' => t('August'),
        '9' => t('September'),
        '10' => t('October'),
        '11' => t('November'),
        '12' => t('December'),
      ],
      '#default_value' => $default_values['month'],
      '#theme_wrappers' => [
        'form_element' => [
          '#attributes' => [
            'class' => ["usa-form-group", "usa-form-group--month", "usa-form-group--select"],
          ],
        ],
      ],
    ];

    $element['day'] = [
      '#type' => 'textfield',
      '#title' => t('Day'),
      '#default_value' => $default_values['day'],
      '#maxlength' => 2,
      '#attributes' => [
        'class' => ['usa-input'],
        'aria-describedby' => "$id-hint",
        'pattern' => '[0-9]*',
        'inputmode' => 'numeric',
      ],
      '#theme_wrappers' => [
        'form_element' => [
          '#attributes' => [
            'class' => ["usa-form-group", "usa-form-group--day"],
          ]
        ],
      ],
    ];

    $element['year'] = [
      '#type' => 'textfield',
      '#title' => t('Year'),
      '#default_value' => $default_values['year'],
      '#maxlength' => 4,
      '#minlength' => 4,
      '#attributes' => [
        'class' => ['usa-input'],
        'aria-describedby' => "$id-hint",
        'pattern' => '[0-9]*',
        'inputmode' => 'numeric',
      ],
      '#theme_wrappers' => [
        'form_element' => [
          '#attributes' => [
            'class' => ["usa-form-group", "usa-form-group--year"],
          ]
        ],
      ],
    ];
    return $element;
  }

  /**
   * Validation callback for this element.
   */
  public static function validateMemorableDate(&$element, FormStateInterface $form_state, &$complete_form) {
    $year = $element['year']['#value'] ?? '';
    $month = $element['month']['#value'] ?? '';
    $day = $element['day']['#value'] ?? '';

    $formatted_date = NULL;
   // Ensure that we either have all three fields, or none.
    if ($year || $month || $day) {
      if (!($year && $month && $day)) {
        if ($element['#required']) {
          $message = t('The date is not complete. Please enter month, day, and year.');
        }
        else {
          $message = t('The date is not complete. Please enter month, day, and year, or clear all fields.');
        }
        $form_state->setError($element, $message);
      }
      elseif (strlen($year) != 4) {
        $form_state->setError($element, t('Please enter a four-digit year.'));
      }
      else {
        $timestamp = mktime(12, 0, 0, $month, $day, $year);
        $padded_month = str_pad($month, 2, '0', STR_PAD_LEFT);
        $padded_day = str_pad($day, 2, '0', STR_PAD_LEFT);
        $formatted_date = "$year-$padded_month-$padded_day";
        $args = ['@m' => $month, '@d' => $day, '@y' => $year];
        if ($formatted_date != date('Y-m-d', $timestamp)) {
          $form_state->setError($element, t('The date you entered (@m/@d/@y) is not valid. Please make sure the day is within the proper bounds for the month you selected.', $args));
        }
        elseif (!empty($element['#min']) && $formatted_date < $element['#min']) {
          $form_state->setError($element, t('The date you entered is earlier than the minimum allowed value of @m/@d/@y.', $args));
        }
        elseif (!empty($element['#max']) && $formatted_date > $element['#max']) {
          $form_state->setError($element, t('The date you entered exceeds the maximum allowed value of @m/@d/@y.', $args));
        }
      }
    }
    $form_state->setValueForElement($element, $formatted_date);
    if (count($form_state->getErrors()) == 0) {
      $form_state->unsetValue($element['month']['#parents'] ?? []);
      $form_state->unsetValue($element['day']['#parents'] ?? []);
      $form_state->unsetValue($element['year']['#parents'] ?? []);
    }
  }

  /**
   * Pre-render callback for this element.
   */
  public static function preRenderMemorableDate(array $element): array {
    $id = $element['#id'];
    $required = $element['#required'];

    $value = $element['#value'];
    $values = ['year' => NULL, 'month' => NULL, 'day' => NULL];
    if (preg_match('/^(?P<year>[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})/', $value, $matches)) {
      $values = array_filter($matches, fn($key) => !is_numeric($key), ARRAY_FILTER_USE_KEY);
    }

    foreach (['month', 'day', 'year'] as $key) {
      $element[$key]['#value'] = $values[$key];
      $element[$key]['#attributes']['aria-describedby'] = "$id-hint";
      $element[$key]['#attributes']['id'] = "$id-$key";
      if ($key == 'month') {
        $element[$key]['#attributes']['class'][] = 'usa-select';
      }
      else {
        $element[$key]['#attributes']['pattern'] = "[0-9]*";
        $element[$key]['#attributes']['class'][] = 'usa-input';
        $element[$key]['#attributes']['inputmode'][] = 'numeric';
      }
      if ($required) {
        $element[$key]['#attributes']['required'] = "required";
        $element[$key]['#attributes']['aria-required'] = "true";
      }
    }

    return $element;
  }
}
