<?php

namespace Drupal\uswds_storybook\Element;

use Drupal\Core\Render\Attribute\FormElement;
use Drupal\Core\Render\Element\Textarea;

/**
 * Provides a form element for multi-line text with a validation checklist.
 *
 * Properties:
 * - #rows: Number of rows in the text box.
 * - #cols: Number of columns in the text box.
 * - #resizable: Controls whether the text area is resizable.  Allowed values
 *   are "none", "vertical", "horizontal", or "both" (defaults to "vertical").
 * - #maxlength: The maximum amount of characters to accept as input.
 * - #validators: Array of validation rules. Each validator is an array with the
 *   following keys:
 *   - id: Text string uniquely identifying this validation rule.
 *   - label: Human-readable (translated) name of the validation rule.
 *   - regex: Javascript-compatible regular expression.
 *
 * Usage example:
 *
 * @code
 * $form['text'] = [
 *   '#type' => 'vaidated_textarea',
 *   '#title' => $this->t('Text'),
 *   '#validators' => [
 *     ['id' => 'numeric', 'label' => $this->t('At least one number'), 'regex' => '[0-9]'],
 *     ['id' => 'uppercase', 'label' => $this->t('At least one uppercase letter'), 'regex' => '[A=-Z]'],
 *   ],
 * ];
 * @endcode
 *
 * @see \Drupal\Core\Render\Element\Textarea
 */
#[FormElement('validated_textarea')]
class ValidatedTextarea extends Textarea {

  public function getInfo() {
    $class = static::class;
    $info = parent::getInfo();
    array_unshift($info['#process'], [$class, 'processValidatedTextarea']);
    $info['#theme_wrappers'] = ['fieldset__validated_text'];
    $info['#validators'] = [];
    return $info;
  }

  /**
   * Process callback: if no validators, fall back to default theme wrappers.
   */
  public static function processValidatedTextarea(array &$element): array {
    if (empty($element['#validators'])) {
      $element['#theme_wrappers'] = ['form_element'];
    }
    return $element;
  }

  /**
   * Prerender callback: Set appropriate attributes.
   */
  public static function preRenderTextarea(array $element): array {
    // If there are no validators, use standard textarea formatter instead.
    if (empty($element['#validators'])) {
      static::setAttributes($element, ['usa-textarea']);
      return $element;
    }
    // Replace form-textarea class with usa-textarea.
    $class =& $element['#attributes']['class'];
    $index = array_search('form-textarea', $class);
    if ($index === FALSE) {
      $index = count($class);
    }
    $class[$index] = 'usa-textarea';

    // Set additional data attributes for validation.
    foreach ($element['#validators'] as $validator) {
      $element['#attributes']['data-validate-' . $validator['id']] = $validator['regex'];
    }
    $element['#attributes']['data-validation-element'] = 'validate-' . $element['#id'];
    return $element;
  }
}
