<?php

namespace Drupal\uswds_storybook\Element;

use Drupal\Core\Render\Attribute\FormElement;
use Drupal\Core\Render\Element;
use Drupal\Core\Render\Element\Textfield;

/**
 * Provides a one-line text field form element with a validation checklist.
 *
 * Properties:
 * - #maxlength: Maximum number of characters of input allowed.
 * - #size: The size of the input element in characters.
 * - #pattern: A string for the native HTML5 pattern attribute.
 * - #validators: Array of validation rules. Each validator is an array with the
 *   following keys:
 *   - id: Text string uniquely identifying this validation rule.
 *   - label: Human-readable (translated) name of the validation rule.
 *   - regex: Javascript-compatible regular expression.
 * Usage example:
 *
 * @code
 * $form['title'] = [
 *   '#type' => 'textfield',
 *   '#title' => $this->t('Subject'),
 *   '#default_value' => $node->title,
 *   '#size' => 60,
 *   '#maxlength' => 128,
 *   '#required' => TRUE,
 *   '#validators' => [
 *     ['id' => 'numeric', 'label' => $this->t('At least one number'), 'regex' => '[0-9]'],
 *     ['id' => 'uppercase', 'label' => $this->t('At least one uppercase letter'), 'regex' => '[A=-Z]'],
 *   ],
 * ];
 * @endcode
 *
 * @see \Drupal\Core\Render\Element
 * @see \Drupal\Core\Render\Element\Textarea
 */
#[FormElement('validated_textfield')]
class ValidatedTextfield extends Textfield {

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = static::class;
    $info = parent::getInfo();
    // Replace first callback with our own.
    $info['#process'][0] = [$class, 'processValidatedTextfield'];
    $info['#theme_wrappers'] = ['fieldset__validated_text'];
    $info['#validators'] = [];

    return $info;
  }

  /**
   * Process callback: if no validators, fall back to default theme wrappers.
   */
  public static function processValidatedTextfield(array &$element): array {
    if (empty($element['#validators'])) {
      $element['#theme_wrappers'] = ['form_element'];
    }
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public static function preRenderTextfield($element) {
    if (empty($element['#validators'])) {
      $element['#attributes']['class'][] = 'usa-input';
      return parent::preRenderTextfield($element);
    }

    $element['#attributes']['type'] = 'text';
    Element::setAttributes($element, ['id', 'name', 'value', 'size', 'maxlength', 'placeholder']);
    static::setAttributes($element, ['usa-input']);

    foreach ($element['#validators'] as $validator) {
      $element['#attributes']['data-validate-' . $validator['id']] = $validator['regex'];
    }
    $element['#attributes']['data-validation-element'] = 'validate-' . $element['#id'];

    return $element;
  }

}
