<?php

namespace Drupal\uswds_storybook\Hook;

use Drupal\Core\Asset\LibraryDiscoveryInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Hook\Attribute\Hook;
use Drupal\Core\Security\TrustedCallbackInterface;
use Drupal\Core\Theme\ThemeManagerInterface;
use Drupal\datetime_range\Plugin\Field\FieldWidget\DateRangeWidgetBase;

/**
 * Theme-related hook implementations for uswds_storybook.
 */
class ThemeHooks implements TrustedCallbackInterface {

  protected string $modulePath;

  public function __construct(
    protected LibraryDiscoveryInterface $libraryDiscovery,
    protected ModuleHandlerInterface $moduleHandler,
    protected ThemeManagerInterface $themeManager,
  ) {
    $this->modulePath = $moduleHandler->getModule('uswds_storybook')->getPath();
  }

  /**
   * Implements hook_theme().
   */
  #[Hook('theme')]
  public function theme(): array {
    $elements = [];
    $template_directory = realpath(__DIR__ . '/../../templates');
    $dh = opendir($template_directory);
    while (($file = readdir($dh)) !== FALSE) {
      if (!str_ends_with($file, '--uswds.html.twig')) {
        [$hook_name] = explode('.', str_replace('-', '_', $file), 2);
        $elements[$hook_name] = [
          'render element' => 'element',
        ];
      }
    }
    closedir($dh);
    return $elements;
  }

  /**
   * Implements hook_library_info_alter().
   */
  #[Hook('library_info_alter')]
  public function libraryInfoAlter(&$libraries, $extension): void {
    if ($extension == 'file') {
      // Swap Drupal core's file library for our own.
      $library = $this->libraryDiscovery->getLibraryByName('uswds_storybook', 'file_input');
      if ($library) {
        self::stringifyLibraryKeys($library['js']);
        self::stringifyLibraryKeys($library['css']);
        // For some reason, LibraryDiscoveryParser flattens CSS; we must
        // re-add the 'theme' layer of the CSS structure.
        $library['css'] = ['theme' => $library['css']];
        $libraries['drupal.file'] = $library;
      }
    }
  }

  /**
   * Implements hook_theme_suggestions_alter().
   */
  #[Hook('theme_suggestions_alter')]
  public function themeSuggestionsAlter(array &$suggestions, array &$variables, $hook): void {
    $template_directory = realpath(__DIR__ . '/../../templates');
    $orig_theme = $variables['theme_hook_original'];
    $hook_dashed = str_replace('_', '-', $hook);
    $dh = opendir($template_directory);
    while (($file = readdir($dh)) !== FALSE) {
      if (
        str_starts_with($file, "$hook_dashed--")
        && str_ends_with($file, '--uswds.html.twig')
      ) {
        $suggestions[] = "{$hook}__uswds";
        if ($hook != $orig_theme) {
          $suggestions[] = $orig_theme . '__uswds';
        }
        if (!empty($variables['type'])) {
          $suggestions[] = $hook . '__' . $variables['type'] . '__uswds';
          if ($hook != $orig_theme) {
            $suggestions[] = $orig_theme . '__' . $variables['type'] . '__uswds';
          }
        }
        break;
      }
    }
    closedir($dh);
  }

  /**
   * Implements hook_theme_registry_alter().
   */
  #[Hook('theme_registry_alter')]
  public function themeRegistryAlter(array &$theme_registry): void {
    $my_path = $this->modulePath;
    $template_directory = "$my_path/templates";
    $dh = opendir($template_directory);
    while (($file = readdir($dh)) !== FALSE) {
      if (str_ends_with($file, '--uswds.html.twig')) {
        $new = substr($file, 0, -10);
        $base = substr($file, 0, -17);
        [$ancestor] = explode('--', $file);

        $new_underscore = str_replace('-', '_', $new);
        $base_underscore = str_replace('-', '_', $base);
        $ancestor_underscore = str_replace('-', '_', $ancestor);

        // Find the most specific existing theme registry entry to modify.
        if (isset($theme_registry[$new_underscore])) {
          $relevant_entry = $new_underscore;
        }
        elseif (isset($theme_regitry[$base_underscore])) {
          $relevant_entry = $base_underscore;
        }
        elseif (isset($theme_regitry[$ancestor_underscore])) {
          $relevant_entry = $ancestor_underscore;
        }
        else {
          // Nope. We'll create a new registry entry.
          $relevant_entry = NULL;
        }

        if (isset($relevant_entry)) {
          $theme_registry[$relevant_entry]['path'] = "$my_path/templates";
          $theme_registry[$relevant_entry]['template'] = $new;
        }
        else {
          $relevant_entry = $new_underscore;
          $preprocess = ['template_preprocess'];
          $theme = $this->themeManager->getActiveTheme()->getName();
          foreach ([$ancestor_underscore, $base_underscore, $new_underscore] as $key) {
            if (function_exists("template_preprocess_$key")) {
              $preprocess[] = "template_preprocess_$key";
            }
            if (function_exists("uswds_storybook_preprocess_$key")) {
              $preprocess[] = "uswds_storybook_preprocess_$key";
            }
            if (function_exists("{$theme}_preprocess_$key")) {
              $preprocess[] = "{$theme}_preprocess_$key";
            }
          }

          $theme_registry[$relevant_entry] = [
            // We make a big assumption here!
            'render element' => 'element',
            'type' => 'module',
            'theme path' => $my_path,
            'template' => $new,
            'path' => "$my_path/templates",
            'preprocess functions' => array_unique($preprocess),
          ];
        }
      }
    }
    closedir($dh);
  }

  /**
   * Implements hook_page_attachments_alter().
   *
   * Adds the importmap script to the HTML head if any components are in use.
   */
  #[Hook('page_attachments_alter')]
  public function pageAttachmentsAlter(array &$attachments): void {
    $js_path = base_path() . $this->modulePath . '/js';
    $imports['imports'] = [
      "utils/active-element" => "$js_path/utils/active-element.js",
      "utils/behavior" => "$js_path/utils/behavior.js",
      "utils/debounce" => "$js_path/utils/debounce.js",
      "utils/focus-trap" => "$js_path/utils/focus-trap.js",
      "utils/is-in-viewport" => "$js_path/utils/is-in-viewport.js",
      "utils/is-ios-device" => "$js_path/utils/is-ios-device.js",
      "utils/sanitizer" => "$js_path/utils/sanitizer.js",
      "utils/scrollbar-width" => "$js_path/utils/scrollbar-width.js",
      "utils/select" => "$js_path/utils/select.js",
      "utils/select-or-matches" => "$js_path/utils/select-or-matches.js",
      "utils/toggle" => "$js_path/utils/toggle.js",
      "utils/toggle-field-mask" => "$js_path/utils/toggle-field-mask.js",
      "utils/toggle-form-input" => "$js_path/utils/toggle-form-input.js",
      "utils/validate-input" => "$js_path/utils/validate-input.js",
      // Note that we only need to define imports for USWDS components that are
      // imported by other components.
      "uswds/accordion" => "$js_path/components/accordion.js",
      "uswds/banner" => "$js_path/components/banner.js",
      "uswds/combo-box" => "$js_path/components/combo-box.js",
      "uswds/date-picker" => "$js_path/components/date-picker.js",
      "uswds/header" => "$js_path/components/header.js",
      "uswds/modal" => "$js_path/components/modal.js",
    ];
    $attachments['#attached']['html_head'][] = [
      [
        '#tag' => 'script',
        '#attributes' => [
          'type' => 'importmap',
        ],
        '#value' => json_encode($imports, JSON_UNESCAPED_SLASHES),
        // Give this a negative weight so it rises to the top.
        '#weight' => -1000,
      ],
      'uswds_storybook_importmap',
    ];
  }

  /**
   * Implements hook_element_info_alter().
   */
  #[Hook('element_info_alter')]
  public function elementInfoAlter(array &$info): void {
    $class = static::class;
    foreach ($info as $key => $value) {
      $camel = self::camelize($key);
      $function = "$class::preRender$camel";
      if (is_callable($function)) {
        $info[$key]['#pre_render'][] = $function;
      }
    }
  }

  /**
   * Implements hook_widget_complete_form_alter().
   */
  #[Hook('field_widget_complete_form_alter')]
  public function fieldWidgetCompleteFormAlter(array &$field_widget_complete_form, FormStateInterface $form_state, array $context): void {
    if ($context['widget'] instanceof DateRangeWidgetBase) {
      $field_widget_complete_form['#theme_wrappers'] = [];
      $field_widget_complete_form['#theme'] = 'field_widget__date_range';
    }
  }

  /**
   * Pre-render callback for radio.
   */
  public static function preRenderRadio(array $element): array {
    self::swapClassName($element, 'form-radio', 'usa-radio__input');
    self::handleDisabled($element);
    return $element;
  }

  /**
   * Pre-render callback for radios.
   */
  public static function preRenderRadios(array $element): array {
    $element['#attributes']['class'] = ['usa-fieldset', 'radios-wrapper'];
    return $element;
  }

  /**
   * Pre-render callback for checkbox.
   */
  public static function preRenderCheckbox(array $element): array {
    self::swapClassName($element, 'form-checkbox', 'usa-checkbox__input');
    self::handleDisabled($element);
    $element['#wrapper_attributes']['class'][] = 'usa-checkbox';
    return $element;
  }

  /**
   * Pre-render callback for checkboxes.
   */
  public static function preRenderCheckboxes(array $element): array {
    $element['#attributes']['class'] = ['usa-fieldset', 'checkboxes-wrapper'];
    return $element;
  }

  /**
   * Pre-render callback for search.
   */
  public static function preRenderSearch(array $element): array {
    self::swapClassName($element, 'form-search', 'usa-search__input');
    self::handleDisabled($element);
    return $element;
  }

  /**
   * Pre-render callback for date.
   */
  public static function preRenderDate(array $element): array {
    if ($element['#attributes']['type'] ?? NULL === 'time') {
      // Let Drupal's standard timepicker handle this.
      return $element;
    }
    // Our datepicker and timepicker are a textfield with Javascript applied.
    self::swapClassName($element, 'form-date', 'usa-input');
    self::handleDisabled($element);

    // The USWDS exemplar has no 'type' attribute. This may or may not be
    // important, but we might as well mimic what works. We remove the 'size'
    // attribute because CSS will control the proper size.
    unset(
      $element['#attributes']['type'],
      $element['#attributes']['step'],
      $element['#attributes']['size'],
    );


    if (!empty($element['#value']) || !empty($element['#default_value'])) {
      $which = !empty($element['#value']) ? '#value' : '#default_value';
      $value = self::normalizeDateValue($element[$which]);
      if (isset($value)) {
        $element['#usa_defaults']['data-default-value'] = self::normalizeDateValue($element[$which]);
      }
    }
    if (!empty($element['#attributes']['min'])) {
      $value = self::normalizeDateValue($element['#attributes']['min']);
      if (isset($value)) {
        $element['#usa_defaults']['data-min-date'] = $value;
      }
      unset($element['#attributes']['min']);
    }
    if (!empty($element['#attributes']['max'])) {
      $value = self::normalizeDateValue($element['#attributes']['max']);
      if (isset($value)) {
        $element['#usa_defaults']['data-max-date'] = $value;
      }
      unset($element['#attributes']['max']);
    }
    $id = $element['#attributes']['id'] ?? NULL;
    if (isset($id)) {
      if (!empty($element['#description'])) {
        // @todo Verify that the following is correct.
        $element['#attributes']['aria-describedby'] = $element['#description']['#id'] . " $id-hint";
      }
      else {
        $element['#attributes']['aria-describedby'] = "$id-hint";
      }
    }

    return $element;
  }

  /**
   * Pre-render callback for textfield.
   */
  public static function preRenderTextfield(array $element): array {
    // Only show maxlength counter for fields that have a magic data attribute.
    if (!empty($element['#attributes']['data-show-max'])) {
      return self::preRenderTextarea($element, 'usa-input');
    }
    self::swapClassName($element, 'form-text', 'usa-input');
    self::handleDisabled($element);
    return $element;
  }

  /**
   * Pre-render callback for textarea.
   */
  public static function preRenderTextarea(array $element, string $class_name = 'usa-textarea'): array {
    self::swapClassName($element, 'form-text', $class_name);
    self::handleDisabled($element);
    // Show maxlength counter for all textarea fields with a maxlength set.
    // @todo Ensure that this won't mangle CKEditor5 fields.
    if (($element['#attributes']['maxlength'] ?? 0) > 0) {
      $maxlength_id = $element['#id'] . '--maxlength';
      $element['#attributes']['data-maxlength-id'] = $maxlength_id;
      $element['#attributes']['data-maxlength-initial-message'] = \Drupal::translation()->formatPlural($element['#attributes']['maxlength'], "You can enter up to 1 character", "You can enter up to @count characters"); /* TRANSLATION */
      $element['#attributes']['class'][] = 'usa-character-count__field';
      $element['#attributes']['aria-describedby'] = $maxlength_id
        . (empty($element['#attributes']['aria-describedby']) ? '' : ' ' . $element['#attributes']['aria-describedby']);
    }

    return $element;
  }

  /**
   * Pre-render callback for button.
   */
  public static function preRenderButton(array $element): array {
    self::swapClassName($element, 'button', 'usa-button');
    if (!empty($element['#button_type']) && in_array($element['#button_type'], ['primary', 'secondary', 'danger'])) {
      $old_class = 'button--' . $element['#button_type'];
      $new_class = 'usa-button--' . ($element['#button_type'] == 'primary' ? 'outline' :$element['#button_type']);
      self::swapClassName($element, $old_class, $new_class);
    }
    self::handleDisabled($element);
    return $element;
  }

  /**
   * Pre-render callback for submit.
   */
  public static function preRenderSubmit(array $element): array {
    return self::preRenderButton($element);
  }

  /**
   * Pre-render callback for managed_file.
   */
  public static function preRenderManagedFile(array $element): array {
    unset($element['upload_button'], $element['remove_button'], $element['display']);

    if (!empty($element['#description'])) {
      $element['description'] = [
        '#type' => 'markup',
        '#markup' => $element['#description'],
        '#weight' => 100,
        '#prefix' => '<div class="usa-hint description">',
        '#suffix' => '</div>',
      ];
      unset($element['#description']);
    }

    return $element;
  }

  /**
   * Pre-render callback for file.
   */
  public static function preRenderFile(array $element): array {
    $element['#attributes']['class'][] = 'usa-file-input';
    return $element;
  }

  /**
   * Helper function to turn a snake_case string into ProperCase.
   *
   * @param string $string
   *   The incoming snake_case string.
   *
   * @return string
   *   The resulting ProperCase string.
   */
  protected static function camelize(string $string): string {
    $camels = array_map('ucfirst', explode('_', $string));
    return implode('', $camels);
  }

  /**
   * If an element is disabled, mark it disabled for ARIA as well.
   *
   * @param array $element
   *   The element to be altered.
   */
  protected static function handleDisabled(array &$element): void {
    if (array_key_exists('disabled', $element['#attributes'])) {
      $element['#attributes']['aria-disabled'] = 'true';
    }
  }

  /**
   * Swap one class name for another.
   *
   * If the old class name was not found, the new one is appended to the class
   * list.
   *
   * @param array $element
   *   The element to be altered.
   * @param string $old_class
   *   The class to be replaced.
   * @param string $new_class
   *   The class with which to replace the old class.
   */
  protected static function swapClassName(array &$element, string $old_class, string $new_class): void {
    if (isset($element['#attributes']['class'])) {
      $index = array_search($old_class, $element['#attributes']['class']);
      if ($index === FALSE) {
        $index = count($element['#attributes']['class']);
      }
    }
    else {
      $index = 0;
    }
    $element['#attributes']['class'][$index] = $new_class;
  }

  /**
   * Helper function to turn int keys in a library descriptor into strings.
   *
   * @param array $items
   *   Array of JS or CSS items with potentially numeric keys.
   */
  protected static function stringifyLibraryKeys(array &$items): void {
    $keys = array_keys($items);
    foreach ($keys as $key) {
      $info = $items[$key];
      if (is_int($key) && !empty($info['data'])) {
        $path = $info['data'];
        // Let LibraryDiscoveryParser rebuild the following.
        unset($items[$key]['data'], $items[$key]['group']);
        // Leading slash ensures that we are relative to Drupal root.
        $items["/$path"] = $items[$key];
        unset($items[$key]);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function trustedCallbacks(): array {
    $methods = [];
    foreach (get_class_methods(self::class) as $method) {
      if (str_starts_with($method, 'preRender')) {
        $methods[] = $method;
      }
    }
    return $methods;
  }

  /**
   * Takes an incoming date value, parses it to an integer, then formats it.
   *
   * @var mixed $value
   *   Either a scalar or a Stringable object.
   *
   * @return string|null
   *   Returns a YYYY-MM-DD formatted date string, or NULL if the incoming value
   *   could not be parsed.
   */
  protected static function normalizeDateValue(mixed $value): ?string {
    if (!is_scalar($value)) {
      if (!is_object($value) || !method_exists($value, '__toString')) {
        user_error('Cannot date-normalize a non-scalar value of type ' . get_debug_type($value), E_USER_WARNING);
        return NULL;
      }
    }
    $value = (string) $value;
    // Test to see if it is already an int. Disallow leading zeroes, allow leading
    // minus-sign.
    if (preg_match('/^-?[1-9][0-9]*$/', $value)) {
      $parsed_value = intval($value);
    }
    else {
      $parsed_value = strtotime($value);
    }
    if ($parsed_value === FALSE || $parsed_value === 0) {
      return NULL;
    }
    return date('Y-m-d', $parsed_value);
  }

}
