<?php

namespace Drupal\uswds_storybook\Plugin\Field\FieldWidget;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Field\Attribute\FieldWidget;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItem;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Drupal\datetime\Plugin\Field\FieldWidget\DateTimeWidgetBase;

/**
 * Plugin implementation of the 'usa_simple_datepicker' widget.
 */
#[FieldWidget(
  id: 'usa_simple_date_picker',
  label: new TranslatableMarkup('USWDS simple date'),
  description: new TranslatableMarkup('Displays a “memorable date” date-picker, and (if the field is configured for time) a simple time picker.'),
  field_types: ['datetime'],
)]
class UsaSimpleDatePicker extends DateTimeWidgetBase {

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings(): array {
    return [
      'increment' => '30',
      'datetime_type' => DateTimeItem::DATETIME_TYPE_DATETIME,
      'timezone' => date_default_timezone_get(),
      'min_year' => NULL,
      'max_year' => NULL,
      'min_hour' => NULL,
      'max_hour' => NULL,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state): array {
    // Wrap all the select elements with a fieldset.
    if (!isset($element['#theme_wrappers']) || !in_array('fieldset', $element['#theme_wrappers'])) {
      $element['#theme_wrappers'][] = 'fieldset';
    }

    $timezone = $this->getFieldSetting('timezone');
    $has_time = $this->getFieldSetting('datetime_type') == DateTimeItem::DATETIME_TYPE_DATETIME;
    $element['date'] = [
      '#type' => 'memorable_date',
      '#default_value' => NULL,
      '#date_timezone' => $timezone,
      '#required' => $element['#required'],
    ];
    if (!empty($element['#min_year']) && preg_match('/^[1-9][0-9]{3}$/', $element['#min_year'])) {
      $element['date']['#min'] = $element['#min_year'] . '-01-01';
    }
    if (!empty($element['#max_year']) && preg_match('/^[1-9][0-9]{3}$/', $element['#max_year'])) {
      $element['date']['#max'] = $element['#max_year'] . '-12-31';
    }
    if ($has_time) {
      $increment = intval($this->getSetting('increment'));
      $element['time'] = [
        '#type' => 'approximate_time',
        '#default_value' => NULL,
        '#date_timezone' => $timezone,
        '#required' => $element['#required'],
        '#increment' => $increment,
      ];
      if (
        !empty($element['#min_hour'])
        && preg_match('/^[0-2][0-9]$/', $element['#min_hour'])
        && ($int_hour = intval($element['#min_hour']))
        && $int_hour >= 0 && $int_hour <= 23
      ) {
        $element['time']['#min'] = $element['#min_hour'] . ':00:00';
      }
      if (
        !empty($element['#max_hour'])
        && preg_match('/^[0-2][0-9]$/', $element['#max_hour'])
        && ($int_hour = intval($element['#max_hour']))
        && $int_hour >= 0 && $int_hour <= 23
      ) {
        $element['time']['#max'] = $element['#max_hour'] . ':59:59';
      }
    }
    if ($items[$delta]->date) {
      // Set the default date (and time, if called for), making sure that it is
      // valid.
      $datetime = $this->createDefaultValue($items[$delta]->date, $timezone);
      $year_int = intval($datetime->format('Y'));
      $month_int = intval($datetime->format('n'));
      $day_int = intval($datetime->format('j'));
      $datetime->setDate($year_int, $month_int, $day_int);

      if ($has_time) {
        $hour_int = intval($datetime->format('G'));
        $minute_int = intval(ltrim($datetime->format('i'), '0'));
        if ($increment == 60) {
          if ($minute_int > 30) {
            // Round up an hour. We will let datetime handle any rollover that
            // happens.
            $hour_int++;
          }
          $minute_int = 0;
        }
        else {
          // Round time to the nearest $increment value.
          $remainder = $minute_int % $increment;
          // If we didn't land on an increment boundary, we need to do some
          // rounding. We will let datetime handle any rollover that happens.
          if ($remainder > 0) {
            $minute_int = round($minute_int / $increment) * $increment;
          }
        }
        $datetime->setTime($hour_int, $minute_int);
      }

      $element['date']['#default_value'] = $datetime->format('Y-m-d');
      if ($has_time) {
        $element['time']['#default_value'] = $datetime->format('H:i') . ':00';
      }
    }

    return $element;

  }

  /**
   * {@inheritdoc}
   */
  public function massageFormValues(array $values, array $form, FormStateInterface $form_state): array {
    $datetime_type = $this->getFieldSetting('datetime_type');
    $timezone = $this->getFieldSetting('timezone');
    $has_time = ($datetime_type === DateTimeItem::DATETIME_TYPE_DATE);
    $datetime_format = $has_time ? DateTimeItemInterface::DATETIME_STORAGE_FORMAT : DateTimeItemInterface::DATE_STORAGE_FORMAT;
    foreach ($values as &$item) {
      if (
        !empty($item['date'])
        && (!$has_time || !empty($item['time']))
      ) {
        $datetime = DrupalDateTime::createFromFormat($datetime_format, $item['date'], $timezone);
        // Because the lowest step granularity available is 5, we can simplify
        // validation of the final digit to either zero or 5.
        if ($has_time && preg_match('/^(2[0-3]|[01][0-9]):([0-5][05])$/', $item['time'], $matches)) {
          $hour = intval(ltrim($matches[1], '0'));
          $minute = intval(ltrim($matches[2], '0'));
          $datetime->setTime($hour, $minute);
        }
        $item['value'] = $datetime->format($datetime_format);
      }
      // Remove extraneous value columns.
      unset($item['date'], $item['time']);
    }
    return $values;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    $this_year = intval(date('Y'));
    $element = [];
    $element['min_year'] =  [
      '#type' => 'number',
      '#title' => t('Minimum year'),
      '#min' => 1000,
      '#max' => 9999,
      '#default_value' => $this->getSetting('min_year') ?? $this_year - 20,
    ];
    $element['max_year'] =  [
      '#type' => 'number',
      '#title' => t('Maximum year'),
      '#min' => 1000,
      '#max' => 9999,
      '#default_value' => $this->getSetting('max_year') ?? $this_year + 20,
      '#element_validate' => [self::class, 'validateMinMaxYear'],
    ];

    if ($this->getFieldSetting('datetime_type') == 'datetime') {
      $element['min_hour'] = [
        '#type' => 'number',
        '#title' => t('Minimum hour'),
        '#default_value' => $this->getSetting('min_hour') ?? 0,
        '#min' => 0,
        '#max' => 23,
      ];
      $element['max_hour'] = [
        '#type' => 'number',
        '#title' => t('Maximum hour'),
        '#default_value' => $this->getSetting('max_hour') ?? 0,
        '#min' => 0,
        '#max' => 23,
        '#element_validate' => [self::class, 'validateMinMaxHour'],
      ];
      $element['increment'] = [
        '#type' => 'radios',
        '#title' => t('Time increment'),
        '#description' => t('The amount of time between options in the select box.'),
        '#options' => [
          5 => t('@count minutes', ['@count' => 5]),
          10 => t('@count minutes', ['@count' => 10]),
          15 => t('@count minutes', ['@count' => 15]),
          20 => t('@count minutes', ['@count' => 20]),
          30 => t('@count minutes', ['@count' => 30]),
          60 => t('1 hour'),
        ],
        '#default_value' => $this->getSetting('increment') ?? 30,
      ];
    }
    return $element;
  }

  /**
   * Validation callback to ensure year max is >= year min, if both are present.
   */
  protected static function validateMinMaxYear(array $element, FormStateInterface $form_state, array $form): void {
    self::validateMinMax(
      $element,
      $form_state,
      'min_year',
      t('Minimum year must be less than or equal to maximum year.'),
    );
  }

  /**
   * Validation callback to ensure hour max is >= hour min if both are present.
   */
  protected static function validateMinMaxHour(array $element, FormStateInterface $form_state, array $form): void {
    self::validateMinMax(
      $element,
      $form_state,
      'min_hour',
      t('Minimum hour must be less than or equal to maximum hour.'),
    );
  }

  /**
   * Helper function to ensure that one field's value is greater than another's.
   *
   * @param array $element
   *   The max form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   State of the form.
   * @param string $min_field_name
   *   The name of the min form element.
   * @param \Drupal\Core\StringTranslation\TranslatableMarkup|string $error_message
   *   The error message to be set on the max element if it fails validation.
   */
  protected static function validateMinMax(
    array $element,
    FormStateInterface $form_state,
    string $min_field_name,
    TranslatableMarkup|string $error_message,
  ): void {
    if (!isset($element['#value'])) {
      // There is no max value, and therefore validity is guaranteed.
      return;
    }
    $max = intval($element['#value']);
    $parents = $element['#parents'];
    // We need to find the sibling field of $element, so we pop up one level.
    array_pop($parents);
    // Now we descend to the name of the field containing the minimum value.
    $parents[] = $min_field_name;
    // Fetch the minimum value if it exists. (If it doesn't, we get NULL back.)
    $min = NestedArray::getValue($form_state->getValues(), $parents);
    if (!isset($min)) {
      // There is no min value, and therefore validity is guaranteed.
      return;
    }
    // Compare max and min, ensuring that min <= max.
    if (intval($min) > $max) {
      $form_state->setError($element, $error_message);
    }
  }

}
