<?php

namespace Drupal\uswds_storybook\Twig;

use Drupal\Component\Render\MarkupInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Menu\MenuLinkTreeElement;
use Drupal\Core\Render\Markup;
use Drupal\Core\Template\Attribute;
use Twig\Extension\AbstractExtension;
use Twig\TwigFilter;
use Twig\TwigFunction;

/**
 * Twig helper extension for USWDS.
 */
class UswdsComponentsTwig extends AbstractExtension {

  /**
   * Build this extension.
   */
  public function __construct(
    protected ModuleHandlerInterface $moduleHandler,
  ) {
  }

  /**
   * {@inheritdoc}
   */
  public function getFunctions() {
    return [
      new TwigFunction('uswds_storybook_asset_path', $this->getComponentAssetPath(...)),
      new TwigFunction('create_attribute', $this->createAttribute(...)),
      new TwigFunction('localized_month_day', $this->formatLocalizedMonthDay(...)),
      new TwigFunction('preprocess_menu', $this->preprocessMenu(...)),
    ];
  }

  public function getFilters() {
    return [
      new TwigFilter('markup_wrapper', $this->toMarkup(...)),
      new TwigFilter('array_replace', $this->arrayReplace(...)),
      new TwigFilter('set_property', $this->setProperty(...)),
      new TwigFilter('localized_long_date', $this->formatLocalizedLongDate(...)),
    ];
  }

  /**
   * Gets the web-accessible path of this module's assets directory.
   *
   * @return string
   */
  public function getComponentAssetPath(): string {
    return base_path() . $this->moduleHandler->getModule('uswds_storybook')->getPath() . '/img';
  }

  /**
   * Takes an array of attributes and returns an Attribute wrapper around them.
   *
   * This is used for Storybook integration.
   *
   * @param array $attributes
   *   Key-value pairs of attributes.
   *
   * @return \Drupal\Core\Template\Attribute
   */
  public function createAttribute(array $attributes = []): Attribute {
    return new Attribute($attributes);
  }

  /**
   * Takes a string and wraps it as Markup.
   *
   * @param string $text
   *   The incoming text
   *
   * @return \Drupal\Component\Render\MarkupInterface|string
   */
  public function toMarkup(string $text): MarkupInterface|string {
    return Markup::create($text);
  }

  /**
   * Finds a string in an array and replaces it with another string.
   *
   * If the string is not found, the new string is appended to the array.
   *
   * @param array $array
   *   The array in which the substitution is to be performed.
   * @param string $find
   *   The string to be replaced in the array.
   * @param string $replace
   *   The string which is to replace the searched-for string.
   *
   * @return array
   *   The resulting array.
   */
  public function arrayReplace(array $array, string $find, string $replace): array {
    $index = array_search($find, $array);
    if ($index === FALSE) {
      $index = count($array);
    }
    $array[$index] = $replace;
    return $array;
  }

  /**
   * Sets a property on an array or object.
   *
   * If the object cannot handle dynamic property assignment, this may fail. In
   * this case, a warning message will be generated, which will be logged (if
   * Drupal logging is enabled).
   *
   * @param array|object $object
   *   The item to which the property is to be assigned.
   * @param string $name
   *   The name of the property to be set.
   * @param mixed $value
   *   The value of the property to be set.
   *
   * @return array|object
   *   The array or object, with the appropriate
   */
  public function setProperty(array|object $object, string $name, mixed $value): array|object {
    if (is_object($object) && !$object instanceof \ArrayAccess) {
      if ($object instanceof \stdClass || property_exists($object, $name) || method_exists($object, '__set')) {
        $object->{$name} = $value;
        return $object;
      }
      try {
        $error_handler = set_error_handler(function(){ }, E_ALL);
        $object->{$name} = $value;
        set_error_handler($error_handler);
        return $object;
      }
      catch (\Exception) {
        user_error("Cannot assign property '$name' to object of type " . get_class($object), E_USER_WARNING);
        return $object;
      }
    }
    else {
      $object[$name] = $value;
      return $object;
    }
  }

  /**
   * Returns a localized long date.
   *
   * If the 'intl' PHP extension is not loaded, we fall back to the Drupal
   * date-formatting service, which is less robust.
   *
   * @param mixed $date
   *   The Unix timestamp of the date to be formatted.
   *
   * @return string
   *   The date formatted according to the current locale.
   */
  public function formatLocalizedLongDate(mixed $date): string {
    $date = intval($date);
    $lang = \Drupal::languageManager()->getCurrentLanguage()->getId();
    // Short-circuit for the happy path.
    if ($lang == 'en' || $lang == 'en_US') {
      return date('F j, Y', $date);
    }
    // Short-circuit for the least happy path.
    if (!extension_loaded('intl')) {
      /** @var \Drupal\Core\Datetime\DateFormatterInterface $formatter */
      $formatter = \Drupal::service('date.formatter');
      return $formatter->format($date, 'custom', 'F j, Y');
    }

    static $intl_formatter;
    if (!isset($intl_formatter)) {
      $timezone = date_default_timezone_get();
      $intl_formatter = new \IntlDateFormatter(
        $lang,
        \IntlDateFormatter::LONG,
        \IntlDateFormatter::NONE,
        $timezone,
        \IntlDateFormatter::GREGORIAN,
      );
    }
    return $intl_formatter->format($date);
  }

  /**
   * Formats a localized month/day combination, in the correct order for locale.
   *
   * For instance, in U.S. English we would say "Apr 23", but in Spanish it is
   * "23 abr". Be aware that not all languages will return the day as purely
   * numeric. For Chinese, for example, it will be "4月 23日".
   *
   * If the intl extension is missing, we return correct translation for the
   * month, but we will always return month and day in the U.S. English order.
   *
   * @param mixed $date
   *   The Unix timestamp to be rendered.
   *
   * @return array
   *   A key/value array, with keys of 'month' and 'day', in the correct order
   *   for the current language.
   */
  public function formatLocalizedMonthDay(mixed $date): array {
    $date = intval($date);
    $lang = \Drupal::languageManager()->getCurrentLanguage()->getId();
    if ($lang == 'en' || $lang == 'en_US') {
      // Short-circuit for the happy path.
      return ['month' => date('M', $date), 'day' => date('j', $date)];
    }
    if (!extension_loaded('intl')) {
      // Fallback option. This will always give the order as month/day even when
      // the current language/locale specifies day/month.
      /** @var \Drupal\Core\Datetime\DateFormatterInterface $formatter */
      $formatter = \Drupal::service('date.formatter');
      $month = $formatter->format($date, 'custom', 'M');
      $day = date('j', $date);
      return ['month' => $month, 'day' => $day];
    }
    static $intl_formatter;
    $timezone = date_default_timezone_get();
    if (!isset($intl_formatter)) {
      $intl_formatter = new \IntlDateFormatter(
        $lang,
        \IntlDateFormatter::MEDIUM,
        \IntlDateFormatter::NONE,
        $timezone,
        \IntlDateFormatter::GREGORIAN,
      );
    }
    $formatted_date = $intl_formatter->format($date);
    // Some locales, such as Arabic ('ar'), return a spaceless string for
    // MEDIUM dates, so we must handle these with LONG instead.
    if (!str_contains($formatted_date, ' ')) {
      $intl_formatter = new \IntlDateFormatter(
        $lang,
        \IntlDateFormatter::LONG,
        \IntlDateFormatter::NONE,
        $timezone,
        \IntlDateFormatter::GREGORIAN,
      );
      $formatted_date = $intl_formatter->format($date);
    }

    $parts = preg_split('/[ ,]+/', $formatted_date);
    // This is assembled piecewise so that we can see the correct sequence in
    // left-to-right format.
    $arabic_numeral_range = '٠' . '-' . '٩';
    // There are other language-specific numbers, but I believe they are mostly
    // of historic interest only.
    // @see https://en.wikipedia.org/wiki/Vai_syllabary#Digits

    foreach ($parts as $i => $part) {
      // Ignore year. Some languages (e.g. Chinese and others) will have the
      // year number embedded with other text, so we do a partial match.
      // In order to catch Persian, we also need to match Arabic numerals.
      if (preg_match("/[0-9$arabic_numeral_range]{4}/u", $part)) {
        unset($parts[$i]);
        break;
      }
    }
    [$part_1, $part_2] = array_values($parts);
    $day = date('j', $date);
    $month = date('n', $date);
    $arabic_day = $arabic_month = '';
    for ($i = 0; $i < strlen($day); $i++) {
      $arabic_day .= mb_chr(intval(substr($day, $i, 1)) + 1632);
    }
    for ($i = 0; $i < strlen($month); $i++) {
      $arabic_month .= mb_chr(intval(substr($month, $i, 1)) + 1632);
    }

    // Handle languages that return numeric prefixes/suffixes to month and day.
    if (
      (self::startsWithNumeric($part_1) && self::startsWithNumeric($part_2))
      || (self::endsWithNumeric($part_1) && self::endsWithNumeric($part_2))
    ) {
      if (
        (
          str_starts_with($part_1, $day)
          || str_starts_with($part_1, $arabic_day)
          || str_ends_with($part_1, $day)
          || str_ends_with($part_1, $arabic_day)
        )
        && (
          str_starts_with($part_2, $month)
          || str_starts_with($part_2, $arabic_month)
          || str_ends_with($part_2, $month)
          || str_ends_with($part_2, $arabic_month)
        )
      ) {
        return ['day' => $part_1, 'month' => $part_2];
      }
      elseif (
        (
          str_starts_with($part_2, $day)
          || str_ends_with($part_2, $day)
          || str_starts_with($part_2, $arabic_day)
          || str_ends_with($part_2, $arabic_day)
        )
        && (
          str_starts_with($part_1, $month)
          || str_ends_with($part_1, $month)
          || str_starts_with($part_1, $arabic_month)
          || str_ends_with($part_1, $arabic_month)
        )
      ) {
        return ['month' => $part_1, 'day' => $part_2];
      }
    }

    if (preg_match("/^[0-9$arabic_numeral_range]{1,2}$/u", $part_1)) {
      return ['day' => $part_1, 'month' => $part_2];
    }
    return ['month' => $part_1, 'day' => $part_2];
  }

  /**
   * Simplifies a menu tree to make theming easier.
   *
   * @param array $elements
   *   An array of menu items.
   *
   * @return array
   *   A simplified nested array suitable for passing to nav_primary or
   *   nav_secondary.
   */
  public function preprocessMenu(array $elements): array {
    $output_items = [];
    foreach ($elements as $menu_element) {
      $item = self::processSubtree($menu_element);
      if ($item) {
        $output_items[] = $item;
      }
    }
    return $output_items;
  }

  /**
   * Helper function invoked by preprocessMenu and recursively by itself.
   *
   * @param \Drupal\Core\Menu\MenuLinkTreeElement $element
   *   The link element whose subtree is to be traversed.
   *
   * @return array|null
   *   The resulting subtree array, or NULL if the element item's link is
   *   disabled.
   */
  protected static function processSubtree(MenuLinkTreeElement $element): ?array {
    if (!$element->link->isEnabled()) {
      return NULL;
    }
    $output = [
      'id' => 'menu-' . $element->link->getMenuName() . '-' . hash('xxh3', $element->link->getPluginId()),
      'subtree' => [],
      'url' => $element->link->getUrlObject()->toUriString(),
      'title' => $element->link->getTitle(),
      'in_active_trail' => $element->inActiveTrail,
    ];
    if (!empty($element->subtree)) {
      foreach ($element->subtree as $sub_element) {
        // Recurse!
        $sub_item = self::processSubtree($sub_element);
        if (!empty($sub_item)) {
          $output['subtree'][] = $sub_item;
        }
      }
    }
    return $output;
  }

  /**
   * Helper function to see if a string starts with digits.
   *
   * It will return TRUE if the string starts with standard numeric digits OR
   * Arabic digits.
   *
   * @param string $string
   *   The string to be tested.
   *
   * @return bool
   *   TRUE if the string starts with a numeral.
   */
  protected static function startsWithNumeric(string $string): bool {
    $codepoint = mb_ord(mb_substr($string, 0, 1));
    return ($codepoint >= 48 && $codepoint <= 57)
      || ($codepoint >= 1632 && $codepoint <= 1641);
  }

  /**
   * Helper function to see if a string ends with digits.
   *
   * It will return TRUE if the string ends with standard numeric digits OR
   * Arabic digits.
   *
   * @param string $string
   *   The string to be tested.
   *
   * @return bool
   *   TRUE if the string ends with a numeral.
   */
  protected static function endsWithNumeric(string $string): bool {
    $codepoint = mb_ord(mb_substr($string, -1));
    return ($codepoint >= 48 && $codepoint <= 57)
      || ($codepoint >= 1632 && $codepoint <= 1641);
  }

}
